#include "cubetransport.h"


/** FUNCTION "det" **/

double det (struct double33 m)
{
  return m.x.x * m.y.y * m.z.z - m.x.x * m.y.z * m.z.y - m.x.y * m.y.x * m.z.z + m.x.y * m.y.z * m.z.x + m.x.z * m.y.x * m.z.y - m.x.z * m.y.y * m.z.x;
}


/** FUNCTION "dist" **/

double dist (struct double3 p0, struct double3 p1)
{
  struct double3 v;
  v.x = p1.x - p0.x;
  v.y = p1.y - p0.y;
  v.z = p1.z - p0.z;
  return sqrt (v.x * v.x + v.y * v.y + v.z * v.z);
}


/** FUNCTION "error" **/

void error (char const *restrict format, ...)
{
  va_list vlist;
  char *buffer;
  int bufsz;

  va_start (vlist, format);
  bufsz = vsnprintf (NULL, 0, format, vlist) + 1;
  buffer = malloc (bufsz);
  vsnprintf (buffer, bufsz, format, vlist);
  va_end (vlist);
  ffprintf (2, stderr, logstream, "ERROR - %s", buffer);
  fprintf (stderr, "Program aborted.\n");
  free (buffer);
  exit (EXIT_FAILURE);
}


/** FUNCTION "ffprintf" **/

void ffprintf (int z, ...)
{
  va_list vlist, thevlist;
  FILE **streams;
  char const *format;
  int i;

  streams = malloc (z * sizeof(FILE *));
  va_start (vlist, z);
  for (i = 0; i < z; ++i)
    {
      streams[i] = va_arg (vlist, FILE *);
    }
  format = va_arg (vlist, char const *);
  for (i = 0; i < z; ++i)
    {
      if (streams[i])
        {
          va_copy (thevlist, vlist);
          vfprintf (streams[i], format, thevlist);
          va_end (thevlist);
        }
    }
  va_end (vlist);
  free (streams);
}


/** FUNCTION "flowat" **/

double flowat (double (***flow)[3], int x, int y, int z, int v)
{
  return okflow (x, y, z, v) ? flow[x][y][z][v] : 0;
}


/** FUNCTION "fopen_ctrl" **/

FILE *fopen_ctrl (char const *restrict filename, char const *restrict mode)
{
  FILE *stream;

  stream = fopen (filename, mode);
  if (!stream)
    {
      error ("I could not open file \"%s\".\n", filename);
    }
  return stream;
}


/** FUNCTION "isround" **/

bool isround (int i)
{
  return i > 1 ? (int) (2 * log (i) / log (CONST_1_618)) > (int) (2 * log (i - 1) / log (CONST_1_618)) : true;
}


/** FUNCTION "lincomb" **/

struct double3 lincomb (double ax, double ay, double az)
{
  struct double3 result;
  result.x = ax * lattice.x.x + ay * lattice.y.x + az * lattice.z.x;
  result.y = ax * lattice.x.y + ay * lattice.y.y + az * lattice.z.y;
  result.z = ax * lattice.x.z + ay * lattice.y.z + az * lattice.z.z;
  return result;
}


/** FUNCTION "okflow" **/

bool okflow (int x, int y, int z, enum dirflow v)
{
  switch (v)
    {
    case X:
      return x >= 0 && x < nnn.x - 1 && y >= 0 && y < nnn.y && z >= 0 && z < nnn.z;
    case Y:
      return x >= 0 && x < nnn.x && y >= 0 && y < nnn.y - 1 && z >= 0 && z < nnn.z;
    case Z:
      return x >= 0 && x < nnn.x && y >= 0 && y < nnn.y && z >= 0 && z < nnn.z - 1;
    default:
      return false;
    }
}


/** FUNCTION "okloop" **/

bool okloop (int x, int y, int z, enum dirloop w)
{
  switch (w)
    {
    case XY:
      return x >= 0 && x < nnn.x - 1 && y >= 0 && y < nnn.y - 1 && z >= 0 && z < nnn.z;
    case XZ:
      return x >= 0 && x < nnn.x - 1 && y >= 0 && y < nnn.y && z >= 0 && z < nnn.z - 1;
    case YZ:
      return x >= 0 && x < nnn.x && y >= 0 && y < nnn.y - 1 && z >= 0 && z < nnn.z - 1;
    default:
      return false;
    }
}


/** FUNCTION "rand1" **/

double rand1 (void)
{
  return (rand () + (rand () + .5) / (RAND_MAX + 1.)) / (RAND_MAX + 1.);
}


/** FUNCTION "scalprod" **/

double scalprod (struct double3 u, struct double3 v)
{
  return u.x * v.x + u.y * v.y + u.z * v.z;
}


/** FUNCTION "strncpy_0" **/

void strncpy_0 (char *restrict dest, const char *restrict src, size_t count)
{
  strncpy (dest, src, count - 1);
  dest [count - 1] = '\0';
}


/** FUNCTION "vectmult" **/

struct double3 vectmult (double a, struct double3 v)
{
  v.x *= a;
  v.y *= a;
  v.z *= a;
  return v;
}


/** FUNCTION "vectsum" */

struct double3 vectsum (struct double3 v, struct double3 w)
{
  v.x += w.x;
  v.y += w.y;
  v.z += w.z;
  return v;
}


/** FUNCTION "warning" **/

void warning (char const *restrict format, ...)
{
  va_list vlist;
  char *buffer;
  int bufsz;

  va_start (vlist, format);
  bufsz = vsnprintf (NULL, 0, format, vlist) + 1;
  buffer = malloc (bufsz);
  vsnprintf (buffer, bufsz, format, vlist);
  va_end (vlist);
  ffprintf (2, stderr, logstream, "ERROR - %s", buffer);
  fprintf (stderr, "Continuing processing...\n");
  free (buffer);
}
