#include "cubetransport.h"
#include "discretized.h"

double bertsekas (int n, struct double3 *start, struct double3 *finish, double costfunc (struct double3 thestart, struct double3 thefinish), double precision)
{


  /** VARIABLE DECLARATIONS **/

  /* Arrays of data */

  double *price = malloc (n * sizeof(double)); // Price of each object
  int *matching = malloc (n * sizeof(int)); // Matching function
  int *gnihctam = malloc (n * sizeof(int)); // Inverse function of "matching"
  double *loss = malloc (n * sizeof(double)); // Loss that each bidder would make with its current object
  int *wishedmate = malloc (n * sizeof(int)); // Wished object of each bidder
  double *wishedloss = malloc (n * sizeof(double)); // Loss corresponding to the wished object of each bidder

  /* Important variables */

  double epsilon; // Bidding increment
  int unhappy; // Bidder which will steal an object
  int stolen; // Bidder which gets stolen
  double cost; // Transportation cost

  /* Indices */

  int i; // Index for the bidders
  int j; // Index for the objects

  /* Auxiliary variables used for recordman-finding subroutines */

  int recordman; // Owner of the record
  double record; // Value of the record
  double score; // Score trying to break the record


  /** INITIALIZATION **/

  /* Matching and inverse matching */

  for (i = 0; i < n; ++i)
    {
      matching[i] = i;
    }

  for (j = 0; j < n; ++j)
    {
      gnihctam[j] = j;
    }

  /* Prices and losses */

  for (j = 0; j < n; ++j)
    {
      price[j] = 0;
    }

  for (i = 0; i < n; ++i)
    {
      loss[i] = costfunc (start[i], finish[matching[i]]) + price[matching[i]];
    }

  /* Wished mates and profits */

  for (i = 0; i < n; ++i)
    {
      recordman = -1;
      record = HUGE_VAL;
      for (j = 0; j < n; ++j)
        {
          score = costfunc (start[i], finish[j]) + price[j];
          if (score < record)
            {
              recordman = j;
              record = score;
            }
        }
      wishedmate[i] = recordman;
      wishedloss[i] = record;
    }

  /* Initial bidding increment */

  epsilon = 0;
  for (i = 0; i < n; ++i)
    {
      epsilon = fmax (epsilon, loss[i] - wishedloss[i]);
    }


  /** PERFORMING AUCTION ALGORITHM **/

  /* Scaling step */

  while (epsilon > precision)
    {
      epsilon /= CONST_2_718;

      while (true) // Start of the auction procedure
        {

          /* We look for an unhappy bidder, and halt the procedure if there is none */

          unhappy = -1;
          for (i = 0; i < n; ++i)
            {
              if (wishedloss[i] < loss[i] - epsilon)
                {
                  unhappy = i;
                  break;
                }
            }
          if (unhappy == -1)
            {
              break;
            }

          /* The unhappy bidder steals the object he wants most and increases its price */

          stolen = gnihctam[wishedmate[unhappy]];
          matching[stolen] = matching[unhappy];
          gnihctam[matching[stolen]] = stolen;
          matching[unhappy] = wishedmate[unhappy];
          gnihctam[matching[unhappy]] = unhappy;
          price[matching[unhappy]] += CONST_0_999 * epsilon;

          /* We update profits for bidders #unhappy and #stolen, resp. wished mates and corresponding profits for other concerned bidders */

          loss[stolen] = costfunc (start[stolen], finish[matching[stolen]]) + price[matching[stolen]];
          loss[unhappy] = costfunc (start[unhappy], finish[matching[unhappy]]) + price[matching[unhappy]];
          for (i = 0; i < n; ++i)
            {
              if (wishedmate[i] == matching[unhappy])
                {
                  recordman = -1;
                  record = HUGE_VAL;
                  for (j = 0; j < n; ++j)
                    {
                      score = costfunc (start[i], finish[j]) + price[j];
                      if (score < record)
                        {
                          recordman = j;
                          record = score;
                        }
                    }
                  wishedmate[i] = recordman;
                  wishedloss[i] = record;
                }
            }
        }
    }


  /** COMPUTING THE TRANSPORTATION COST **/

  cost = 0;
  for (i = 0; i < n; ++i)
    {
      cost += costfunc (start[i], finish[matching[i]]);
    }
  cost /= n;


  /** EXITING THE FUNCTION **/

  free (wishedloss);
  free (wishedmate);
  free (loss);
  free (gnihctam);
  free (matching);
  free (price);
  return cost;
}
