#include "cubetransport.h"
#include "quadratic.h"

void codescent (double *p_flow[], double *p_chargea[], double *p_chargeb[], struct double3 dipolea[], struct double3 dipoleb[], enum dirflow v)
{


  /** VARIABLE DECLARATIONS **/

  /* Data arrays */

  double chargema[T1_MAX], chargemb[T1_MAX]; // Charge at A, resp. at B, along time, at half-step times
  double dflow[T1_MAX]; // Delta of the flow between A and B
  double ds[T1_MAX + 1]; // Delta of the cumulated flow between A and B

  /* Important variables */

  struct double3 e; // Vector from A to B
  double dcost; // Derivative of the cost along the descent direction (made homogeneous to a cost)
  double ddcost; // Second derivative of the cost along the descent direction (made homogeneous to a cost)
  double amplitude, critical; // Length (normalized) along which we will follow the descent direction; and value of amplitude at which singularity would occur

  /* Auxiliary variables */

  int t; // Index for time
  double partial; // Partial derivative of the cost w.r.t. the cumulated flow
  double coeffpre; // Coefficient used for preconditioning
  double dchargem; // Delta of an entry of chargemb (which is also the -Delta of the corresponding entry of chargema)
  struct double3 ddipole; // Delta of an entry of dipolea (which is also the -Delta of the corresponding entry of dipoleb)
  double dsm; // Delta of the cumulated flow at some half-step time
  double thedouble; // Auxiliary double variable representing a dimensionless quantity
  double thecharge; // Auxiliary variable representing a charge
  struct double3 thevect; // Auxiliary vector


  /** SETTING USEFUL VARIABLES **/

  e = lincomb (v == X, v == Y, v == Z);
  for (t = 0; t < t1; ++t)
    {
      chargema[t] = (*p_chargea[t] + *p_chargea[t + 1]) / 2;
      chargemb[t] = (*p_chargeb[t] + *p_chargeb[t + 1]) / 2;
    }


  /** COMPUTING THE DESCENT DIRECTION **/

  /* Partial derivative of the cost */

  dcost = 0; // We will also compute dcost
  for (t = 1; t < t1; ++t)
    {
      partial = 0;
      partial += scalprod (dipolea[t - 1], e) / chargema[t - 1];
      partial += scalprod (dipoleb[t - 1], e) / chargemb[t - 1];
      partial -= scalprod (dipolea[t], e) / chargema[t];
      partial -= scalprod (dipoleb[t], e) / chargemb[t];
      partial += scalprod (dipolea[t - 1], dipolea[t - 1]) / (chargema[t - 1] * chargema[t - 1]) / 2;
      partial -= scalprod (dipoleb[t - 1], dipoleb[t - 1]) / (chargemb[t - 1] * chargemb[t - 1]) / 2;
      partial += scalprod (dipolea[t], dipolea[t]) / (chargema[t] * chargema[t]) / 2;
      partial -= scalprod (dipoleb[t], dipoleb[t]) / (chargemb[t] * chargemb[t]) / 2;

      /* We determine ds thanks to coeffpre, and dcost accordingly */

      coeffpre = fmin (fmin (chargema[t - 1], chargema[t]), fmin (chargemb[t - 1], chargemb[t]));
      ds[t] = coeffpre * partial;
      dcost += partial * ds[t];
    }
  ds[0] = 0;
  ds[t1] = 0;


  /** CHOOSING HOW LONG WE WILL FOLLOW THE DESCENT DIRECTION **/

  /* Computing ddcost */

  ddcost = 0;
  for (t = 0; t < t1; ++t)
    {
      dflow[t] = ds[t + 1] - ds[t]; // Delta of the flow itself, which will be re-used later
      dchargem = (ds[t] + ds[t + 1]) / 2;
      ddipole = vectmult (dflow[t] / 2, e);
      thevect = vectsum (vectmult (chargema[t], ddipole), vectmult (dchargem, dipolea[t]));
      ddcost += 2 * scalprod (thevect, thevect) / (chargema[t] * chargema[t] * chargema[t]);
      thevect = vectsum (vectmult (chargemb[t], ddipole), vectmult (-dchargem, dipoleb[t]));
      ddcost += 2 * scalprod (thevect, thevect) / (chargemb[t] * chargemb[t] * chargemb[t]);
    }

  /* Computing the critical amplitude */

  critical = HUGE_VAL;
  for (t = 0; t < t1; ++t)
    {
      dsm = (ds[t] + ds[t + 1]) / 2;
      if (dsm > 0)
        {
          critical = fmin (critical, chargemb[t] / dsm);
        }
      else if (dsm < 0)
        {
          critical = fmin (critical, chargema[t] / -dsm);
        }
    }

  /* Computing the amplitude of descent */

  thedouble = sqrt (1 + 2 / critical * dcost / ddcost);
  amplitude = 2 * dcost / ddcost / thedouble / (1 + thedouble);


  /** UPDATING "flow", "chargea" AND "chargeb" **/

  for (t = 0; t < t1; ++t)
    {
      *p_flow[t] -= amplitude * dflow[t];
    }
  for (t = 1; t < t1; ++t)
    {
      thecharge = amplitude * ds[t];
      *p_chargea[t] += thecharge;
      *p_chargeb[t] -= thecharge;
    }
}
