/** This is the general header for program CUBETRANSPORT **/

#ifndef CUBETRANSPORT_H_INCLUDED
#define CUBETRANSPORT_H_INCLUDED


/** CUSTOMIZABLE PARAMETERS **/

#define VERSION "1.0" // Version number of the program
#define MODENAME_MAX 33 // Maximum possible length of transportation mode name (including terminating \0)
#define DISBALANCEABS_MAX 3.1E-2 // Tolerable total charge absolute difference between the two .cube files
#define DISBALANCEREL_MAX 3.9E-3 // Tolerable total charge relative difference between the two .cube files
#define NSAM_DEFAULT 512 // Default number of sampling points for discretization
#define PRECISION_BERTSEKAS_DEFAULT 6.1E-5 // Default precision for the Bertsekas algorithm
#define NMC_DEFAULT 8 // Default number of Monte Carlo assessments for estimating the true transportation distance
#define T1_DEFAULT 16 // Default number of time steps for time-continuous transportation
#define T1_MAX 256 // Maximum number of time steps allowed for time-continuous transportation
#define ITERATION_MIN 26 // Minimum number of iterations before declaring stationarity
#define ZERO_DESCENT_DEFAULT 7.6E-6 // Precision level required to stop the gradient descent


/** REQUIRED HEADERS **/

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include <string.h>


/** STRUCTURES **/

/* "int3" is a struct type combining three int variables "x", "y", "z" */

struct int3
{
  int x;
  int y;
  int z;
};

/* "double3" is a struct type combining three double variables "x", "y", "z" */

struct double3
{
  double x;
  double y;
  double z;
};

/* "double33" is a struct type combining three (struct double3) variables "x", "y", "z" */

struct double33
{
  struct double3 x;
  struct double3 y;
  struct double3 z;
};

/* "pointcharge" is a struct type describing a point charge "charge" at point "point" */

struct pointcharge
{
  double charge;
  struct double3 point;
};


/** ENUMERATIONS **/

/* "transmode" is an enum type to indicate the transportation mode */

enum transmode
{
  TV, // Total variation
  KR, // Kantorovich-Rubinstein
  QW, // Quadratic Wasserstein
  LQ, // Linearized quadratic Wasserstein
};

/* "dirflow" is an enum type to indicate the direction of a charge flow */

enum dirflow
{
  X, // Along the (x) direction
  Y, // Along the (y) direction
  Z, // Along the (z) direction
};

/* "dirloop" is an enum type to indicate the direction of a charge flow loop */

enum dirloop
{
  XY, // Along the (xy) plane
  XZ, // Along the (xz) plane
  YZ, // Along the (yz) plane
};


/** GLOBAL VARIABLES **/

enum transmode mode; // Inner encoding of the transportation mode
char modename[MODENAME_MAX]; // Full name of the transportation mode
double precision; // Precision level for the computations
FILE *logstream; // Stream on which the log is written
struct int3 nnn; // Meta-size of the grid
struct double33 lattice; // Dimensions of the lattice cells


/** CORE FUNCTION PROTOTYPES **/

/* Main function has the following syntax: ./cubetransport file0.cube file1.cube [-m=mode] [-p=precision] [-l=logfile] */

int main (int argc, char *argv[]);

/* Function "recube" changes punctual density data into volumic data. "*p_density" is the place data are stocked at and "nnn" is the size of the data grid */

void recube (double ****p_density, struct int3 nnn);

/* Functions "tvdistance", resp. "krdistance", etc. compute the transportation distances for mode TV, resp. KR, etc.. Data are stored into meta-arrays "charge0" and "charge1"; the other parameters are passed as global variables */

double tvdistance (double ***charge0, double ***charge1);
double krdistance (double ***charge0, double ***charge1);
double qwdistance (double ***charge0, double ***charge1);
double lqdistance (double ***charge0, double ***charge1);


/** AUXILIARY FUNCTION PROTOTYPES **/

#include "auxfunctions.h"

#endif // CUBETRANSPORT_H_INCLUDED
