#include "cubetransport.h"
#include "quadratic.h"

void descent (double (***flow)[3], double ***charge0, double ***charge1)
{


  /** VARIABLE DECLARATIONS **/

  /* Meta-arrays */

  double ***chargem; // Mean charges
  struct double3 ***move; // Vectors by which charge moves
  double (***dloop)[3]; // Delta of loop coefficients
  double (***dflow)[3]; // Delta of the flow

  /* Important variables */

  double dcost; // Derivative of the cost along the descent direction (made homogeneous to a cost)
  double ddcost; // Second derivative of the cost along the descent direction (made homogeneous to a cost)
  double amplitude; // Length (normalized) that we will follow the descent direction

  /* Auxiliary variables */

  int x, y, z; // Indices for locating a node in the grid
  enum dirflow v; // Index for a flow direction
  enum dirloop w; // Index for a loop direction
  double partial; // Partial derivative of the cost w.r.t. the loop coefficient
  double coeffpre; // Coefficient used for preconditioning
  double fx, fy, fz; // Transported charge along the x direction, resp. along y, resp. along z
  double theloop; // Auxiliary variable representing a loop coefficient
  struct double3 thevect; // Auxiliary vector


  /** PREPARING STUFF **/

  /* Allocating memory */

  METAMALLOC (chargem, nnn);
  METAMALLOC (move, nnn);
  METAMALLOC (dflow, nnn);
  METAMALLOC (dloop, nnn);

  /* Computing "chargem" */

  METAFOR (x, y, z, nnn)
  {
    chargem[x][y][z] = (charge0[x][y][z] + charge1[x][y][z]) / 2;

    /* Computing "move" */

    fx = (flowat (flow, x - 1, y, z, X) + flowat (flow, x, y, z, X)) / 2;
    fy = (flowat (flow, x, y - 1, z, Y) + flowat (flow, x, y, z, Y)) / 2;
    fz = (flowat (flow, x, y, z - 1, Z) + flowat (flow, x, y, z, Z)) / 2;
    thevect = lincomb (fx, fy, fz);
    move[x][y][z] = vectmult (1 / chargem[x][y][z], thevect);

    /* Setting "dflow" to 0 */

    for (v = 0; v < 3; ++v)
      {
        dflow[x][y][z][v] = 0;
      }
  }
  ENDMETAFOR


  /** COMPUTING THE DESCENT DIRECTION **/

  dcost = 0; // We will also compute dcost
  METAFOR (x, y, z, nnn) // We scan all the valid loops
  {
    for (w = 0; w < 3; ++w)
      {
        if (okloop (x, y, z, w))
          {

            /* Computing the partial derivative of the cost functional along the loop coefficient; and setting a relevant coefficient for preconditioning */

            partial = 0;
            switch (w)
              {
              case XY: // Case of a loop in the (xy) direction
                partial += scalprod (lincomb (+1, -1, 0), move[x][y][z]);
                partial += scalprod (lincomb (+1, +1, 0), move[x + 1][y][z]);
                partial += scalprod (lincomb (-1, +1, 0), move[x + 1][y + 1][z]);
                partial += scalprod (lincomb (-1, -1, 0), move[x][y + 1][z]);
                coeffpre = chargem[x][y][z] + chargem[x][y + 1][z] + chargem[x + 1][y][z] + chargem[x + 1][y + 1][z];
                break;
              case XZ: // Case of a loop in the (xz) direction
                partial += scalprod (lincomb (+1, 0, -1), move[x][y][z]);
                partial += scalprod (lincomb (+1, 0, +1), move[x + 1][y][z]);
                partial += scalprod (lincomb (-1, 0, +1), move[x + 1][y][z + 1]);
                partial += scalprod (lincomb (-1, 0, -1), move[x][y][z + 1]);
                coeffpre = chargem[x][y][z] + chargem[x][y][z + 1] + chargem[x + 1][y][z] + chargem[x + 1][y][z + 1];
                break;
              case YZ: // Case of a loop in the (yz) direction
                partial += scalprod (lincomb (0, +1, -1), move[x][y][z]);
                partial += scalprod (lincomb (0, +1, +1), move[x][y + 1][z]);
                partial += scalprod (lincomb (0, -1, +1), move[x][y + 1][z + 1]);
                partial += scalprod (lincomb (0, -1, -1), move[x][y][z + 1]);
                coeffpre = chargem[x][y][z] + chargem[x][y][z + 1] + chargem[x][y + 1][z] + chargem[x][y + 1][z + 1];
                break;
              }

            /* We determine dloop thanks to coeffpre, and dcost accordingly */

            dloop[x][y][z][w] = coeffpre * partial;
            dcost += partial * dloop[x][y][z][w];
          }
      }
  }
  ENDMETAFOR

  /* Converting the field of loops into a flow */

  METAFOR (x, y, z, nnn)
  {
    for (w = 0; w < 3; ++w)
      {
        if (okloop (x, y, z, w))
          {
            theloop = dloop[x][y][z][w];
            switch (w)
              {
              case XY:
                dflow[x][y][z][X] += theloop;
                dflow[x + 1][y][z][Y] += theloop;
                dflow[x][y + 1][z][X] -= theloop;
                dflow[x][y][z][Y] -= theloop;
                break;
              case XZ:
                dflow[x][y][z][X] += theloop;
                dflow[x + 1][y][z][Z] += theloop;
                dflow[x][y][z + 1][X] -= theloop;
                dflow[x][y][z][Z] -= theloop;
                break;
              case YZ:
                dflow[x][y][z][Y] += theloop;
                dflow[x][y + 1][z][Z] += theloop;
                dflow[x][y][z + 1][Y] -= theloop;
                dflow[x][y][z][Z] -= theloop;
                break;
              }
          }
      }
  }
  ENDMETAFOR


  /** CHOOSING HOW LONG WE WILL FOLLOW THE DESCENT DIRECTION **/

  /* Computing the second derivative of the cost functional along the descent direction */

  ddcost = 0;
  METAFOR (x, y, z, nnn)
  {
    fx = (flowat (dflow, x - 1, y, z, X) + flowat (dflow, x, y, z, X)) / 2;
    fy = (flowat (dflow, x, y - 1, z, Y) + flowat (dflow, x, y, z, Y)) / 2;
    fz = (flowat (dflow, x, y, z - 1, Z) + flowat (dflow, x, y, z, Z)) / 2;
    thevect = lincomb (fx, fy, fz);
    ddcost += 2 * scalprod (thevect, thevect) / chargem[x][y][z];
  }
  ENDMETAFOR

  /* Computing the amplitude of descent */

  amplitude = dcost / ddcost;

  /* Updating "flow" */

  METAFOR (x, y, z, nnn)
  {
    for (v = 0; v < 3; ++v)
      {
        if (okflow (x, y, z, v))
          {
            flow[x][y][z][v] -= amplitude * dflow[x][y][z][v];
          }
      }
  }
  ENDMETAFOR


  /** EXITING THE FUNCTION **/

  METAFREE (nnn, dflow);
  METAFREE (nnn, dloop);
  METAFREE (nnn, move);
  METAFREE (nnn, chargem);
}
