#include "cubetransport.h"
#include "discretized.h"

double krdistance (double ***charge0, double ***charge1)
{


  /** VARIABLE DECLARATIONS **/

  /* Parameter-like variables */

  int nsama, nsamb; // Number of sample points for resp. step A and step B of the computation
  int nmca, nmcb; // Number of Monte Carlo assessments for resp. step A and step B of the computation
  double precision_bertsekas; // Precision for the Bertsekas algorithm

  /* Important variables */

  int nnodes; // Number of nodes in the grid
  double totcharge; // Total charge to be transported
  struct pointcharge *loss, *gain; // Charge loss, resp. charge gain measures, as sets of pointcharges
  struct double3 *losssample, *gainsample; // Subsamples of resp. "loss" and "gain", as sets of points all having the same weight
  double distance; // Kantorovich-Rubinstein distance

  /* Variables for aggregating the results */

  double mcmeana, mcmeanb; // Mean of the values of the Monte Carlo loop, resp. for step A and for step B
  double mcvarb; // Variance of the values of the Monte Carlo loop (for step B)
  double coeffextra; // Coefficient used for extrapolation
  double bias; // Discretization bias
  double variancea, varianceb, variance; // Variance of the Monte Carlo estimator, resp. for step A, step B, and global estimation

  /* Auxiliary variables */

  int i; // Generic index
  int x, y, z; // Indices for locating a node in the grid
  struct double3 thepoint; // Auxiliary variable representing a point
  double thecharge; // Auxiliary variable representing a charge
  double mcsum; // Sum of the values of the Monte Carlo loop
  double mcsum2; // Sum of the squares of the values of the Monte Carlo loop
  int progress, progress_prev; // Variables to take the user informed of the progress of computation


  /** DEFINITION OF THE PARAMETERS "nsama", "nsamb", "nmca" AND "nmcb" **/

  /* Setting the parameters */

  nsama = ceil (NSAM_DEFAULT * pow (2, precision / CONST_2_833));
  nmca = ceil (NMC_DEFAULT * pow (2, precision / CONST_8_500));
  nsamb = ceil (nsama / CONST_2_828);
  nmcb = ceil (CONST_8_315 * nmca);
  precision_bertsekas = PRECISION_BERTSEKAS_DEFAULT / pow (2, precision);

  /* Telling the parameters in the log */

  fprintf (logstream, "In main computation, the number of points for the discretization procedure will be %d.\n", nsama);
  fprintf (logstream, "In main computation, the number of Monte Carlo iterations for estimating the transportation distance will be %d.\n", nmca);
  fprintf (logstream, "In extrapolation step, the number of points for the discretization procedure will be %d.\n", nsamb);
  fprintf (logstream, "In extrapolation step, the number of Monte Carlo iterations for estimating the transportation distance will be %d.\n", nmcb);
  fprintf (logstream, "The precision in the Bertsekas algorithm will be %.2E a.u..\n", precision_bertsekas);


  /** FILLING "loss" AND "gain" **/

  /* Un-normalized charges */

  nnodes = nnn.x * nnn.y * nnn.z;
  loss = malloc (nnodes * sizeof(struct pointcharge));
  gain = malloc (nnodes * sizeof(struct pointcharge));
  i = 0;
  METAFOR (x, y, z, nnn)
  {
    thepoint = lincomb (x, y, z);
    loss[i].point = thepoint;
    gain[i].point = thepoint;
    thecharge = charge1[x][y][z] - charge0[x][y][z];
    if (thecharge < 0)
      {
        loss[i].charge = -thecharge;
        gain[i].charge = 0;
      }
    else
      {
        loss[i].charge = 0;
        gain[i].charge = thecharge;
      }
    totcharge += fabs (thecharge);
    ++i;
  }
  ENDMETAFOR
  totcharge /= 2;

  /* Normalizing the charges */

  for (i = 0; i < nnodes; ++i)
    {
      loss[i].charge /= totcharge;
      gain[i].charge /= totcharge;
    }


  /** STEP A: MAIN MONTE CARLO LOOP **/

  /* Preparation steps*/

  progress = 0; // Progress of computation (as a percentage, rounded to the next integer below)
  ffprintf (2, stdout, logstream, "Starting main Monte Carlo loop...\n");
  mcsum = 0;
  losssample = malloc (nsama * sizeof(struct double3));
  gainsample = malloc (nsama * sizeof(struct double3));
  for (i = 0; i < nmca; ++i) // Start of the Monte Carlo loop itself
    {

      /* Displaying the progress of computation */

      progress_prev = progress;
      progress = 100 * CONST_0_618 * i / nmca;
      if (progress > progress_prev)
        {
          printf ("%2d %% done...\n", progress);
        }

      /* Performing a Monte Carlo assessment */

      sample (nnodes, loss, nsama, losssample);
      sample (nnodes, gain, nsama, gainsample);
      distance = totcharge * bertsekas (nsama, losssample, gainsample, dist, precision_bertsekas);
      fprintf (logstream, "Monte Carlo simulation #%d gave a transportation distance of %f a.u..\n", i, distance);
      fflush (logstream);
      mcsum += distance;
    }

  /* Monte Carlo estimator */

  mcmeana = mcsum / nmca;
  fprintf (logstream, "The main Monte Carlo loop estimated the %s distance to be %f a.u..\n", modename, mcmeana);
  fprintf (logstream, "Confidence interval for this estimation will be given later.\n");
  fflush (logstream);


  /** STEP B: EXTRAPOLATION MONTE CARLO LOOP **/

  /* Preparation steps */

  ffprintf (2, stdout, logstream, "Starting extrapolation Monte Carlo loop...\n");
  mcsum = 0;
  mcsum2 = 0;
  losssample = realloc (losssample, nsamb * sizeof(struct double3));
  gainsample = realloc (gainsample, nsamb * sizeof(struct double3));
  for (i = 0; i < nmcb; ++i) // Start of the Monte Carlo loop itself
    {

      /* Displaying the progress of computation */

      progress_prev = progress;
      progress = 100 * (CONST_0_618 + (1 - CONST_0_618) * i / nmcb);
      if (progress > progress_prev)
        {
          printf ("%2d %% done...\n", progress);
        }

      /* Performing a Monte Carlo assessment */

      sample (nnodes, loss, nsamb, losssample);
      sample (nnodes, gain, nsamb, gainsample);
      distance = totcharge * bertsekas (nsamb, losssample, gainsample, dist, precision_bertsekas);
      fprintf (logstream, "Monte Carlo simulation #%d gave a transportation distance of %f a.u..\n", i, distance);
      fflush (logstream);
      mcsum += distance;
      mcsum2 += distance * distance;
    }
  printf ("Done.\n");

  /* Monte Carlo estimator */

  mcmeanb = mcsum / nmcb;
  mcvarb = mcsum2 / nmcb - mcmeanb * mcmeanb;
  varianceb = mcvarb / nmcb;
  variancea = mcvarb * nsamb / nsama / nmca;
  fprintf (logstream, "The extrapolation Monte Carlo loop estimated the %s distance to be %f a.u..\n", modename, mcmeana);
  fprintf (logstream, "The Monte Carlo error bar for the extrapolation step is +-%.2E a.u..\n", CONST_1_959 * sqrt (varianceb));
  fprintf (logstream, "Hence, the Monte Carlo error bar for the main step was +-%.2E a.u..\n", CONST_1_959 * sqrt (variancea));

  /* Freeing memory which is no more needed */

  free (gainsample);
  free (losssample);
  free (gain);
  free (loss);


  /** EXPLOITING THE RESULTS BY RICHARDSON INTERPOLATION **/

  coeffextra = pow (nsamb, CONST_0_666) / (pow (nsama, CONST_0_666) - pow (nsamb, CONST_0_666));
  bias = coeffextra * (mcmeanb - mcmeana);
  distance = mcmeana - bias;
  fprintf (logstream, "This gives a bias-free estimator of the Kantorovitch-Rubinstein distance of %f a.u..\n", distance);
  variance = (coeffextra + 1) * (coeffextra + 1) * variancea + coeffextra * coeffextra * varianceb;
  fprintf (logstream, "The Monte Carlo error for that estimation is about +-%.2E a.u..\n", CONST_1_959 * sqrt (variance));
  fprintf (logstream, "The bias was about %+.3E a.u.  It would be more cautious not to consider it to be fully annihilated, and to take it into account in the margin of error. That would yield a margin of error of +-%.2E a.u.\n", bias, CONST_1_959 * sqrt (variance) + fabs (bias));
  return distance;
}
