/**
    This is the main source file of program CUBETRANSPORT, version 1.0 dated from 2015-06-24.
    Program CUBETRANSPORT and its source files are licensed under the CC-BY-SA 3.0 Unported License. Namely, you are free to share and to remix this work, under the conditions of attributing the work to its author and distributing the work resulting of an alteration under a compatible license.
    Program CUBETRANSPORT was written by R&eacute;mi Peyre (Universit&eacute; de Lorraine - CNRS)
    For documentation or reporting bugs, see http://www.phare.normalesup.org/~rpeyre/pro/research/cubetransport/
**/

#include "cubetransport.h"

int main (int argc, char *argv[])
{


  /** VARIABLE DECLARATIONS **/

  /* Important variables */

  char cube0name[FILENAME_MAX], cube1name[FILENAME_MAX], logname[FILENAME_MAX]; // Names of resp. the first .cube file, the second .cube file and the log file
  FILE *cube0stream, *cube1stream; // Streams corresponding to resp. the first .cube file and the second .cube file
  double ***charge0, ***charge1; // Meta-arrays of charges for resp. the first and the second electronic densities
  double distance; // Transportation distance computed

  /* Less important variables */

  bool test0, test1, testm, testp, testl; // Tests for whether the first .cube file name, resp. the second .cube file name, resp. the transportation mode, etc., were given
  bool testd0n, testd0z, testd1z, testd1n, testh, testn, testr0, testr1; // Miscellaneous tests
  int natoms0, natoms1; // Number of atoms in the first, resp. the second .cube file
  struct int3 nnn1; // Size of the grid for the second .cube file
  double cellvol; // Volume of a grid cell
  double totcharge0, totcharge1, totcharge; // Total charge for the first density, resp. for the second density, resp. mean total charge
  double disbalanceabs; // Absolute disbalance between total charges
  double disbalancerel; // Relative disbalance between total charges

  /* Auxiliary variables */

  int i; // Generic index
  int x, y, z; // Indices for locating a node in the grid
  char *theptr; // Auxiliary pointer to char, used to read input
  char thestring[BUFSIZ], thestring0[BUFSIZ], thestring1[BUFSIZ]; // Auxiliary strings
  double thedensity0, thedensity1; // Auxiliary variables representing densities
  double thecharge0, thecharge1; // Auxiliary variables representing charges
  int theint; // Auxiliary int variable


  /** READING INPUT **/

  /* Default parameters */

  mode = KR;
  precision = 0;
  logstream = NULL;
  strncpy (logname, "cubetransport.log", FILENAME_MAX);

  /* Initializing which command arguments were given */

  test0 = false;
  test1 = false;
  testm = false;
  testp = false;
  testl = false;
  for (i = 1; i < argc; ++i) // Starting to read the input arguments
    {
      theptr = argv[i];
      if (!strncmp (theptr, "-", 1)) // Case the input argument is for an option
        {
          theptr += 1;

          /* Reading an argument for specifying transportation mode */

          if (!strncmp (theptr, "m=", 2))
            {
              if (testm)
                {
                  error ("You have given several mode arguments.\n");
                }
              testm = true;
              theptr += 2;
              if (!strcmp (theptr, "tv"))
                {
                  mode = TV;
                }
              else if (!strcmp (theptr, "kr"))
                {
                  mode = KR;
                }
              else if (!strcmp (theptr, "lq"))
                {
                  mode = LQ;
                }
              else if (!strcmp (theptr, "qw"))
                {
                  mode = QW;
                }
              else
                {
                  error ("I do not know the transportation mode \"%s\".\n", theptr);
                }
            }

          /* Reading an argument for specifying computation precision */

          else if (!strncmp (theptr, "p=", 2))
            {
              if (testp)
                {
                  error ("You have given several precision arguments.\n");
                }
              testp = true;
              theptr += 2;
              theint = sscanf (theptr, "%lf", &precision);
              if (theint < 1)
                {
                  error ("I could not read your precision argument \"%s\".\n", theptr);
                }
            }

          /* Reading an argument for specifying the log file name */

          else if (!strncmp (theptr, "l=", 2))
            {
              if (testl)
                {
                  error ("You have given several logfile arguments.\n");
                }
              testl = true;
              theptr += 2;
              strncpy (logname, theptr, FILENAME_MAX);
            }

          /* Case of an unknown option */

          else
            {
              error ("I do not know the option \"-%s\"", theptr);
            }
        }

      /* Reading an argument for specifying an input file */

      else
        {
          if (!test0)
            {
              test0 = true;
              strncpy (cube0name, theptr, FILENAME_MAX);
            }
          else if (!test1)
            {
              test1 = true;
              strncpy (cube1name, theptr, FILENAME_MAX);
            }
          else
            {
              error ("You have given more than two .cube files.\n");
            }
        }
    }

  /* Looking if some mandatory argument is missing */

  if (!test1)
    {
      error ("Missing .cube file argument.\n");
    }

  /* Giving the transportation mode name */

  switch (mode)
    {
    case TV:
      strncpy (modename, "total variation", MODENAME_MAX - 1);
      break;
    case KR:
      strncpy (modename, "Kantorovich-Rubinstein", MODENAME_MAX - 1);
      break;
    case QW:
      strncpy (modename, "quadratic Wasserstein", MODENAME_MAX - 1);
      break;
    case LQ:
      strncpy (modename, "linearized quadratic Wasserstein", MODENAME_MAX - 1);
      break;
    }

  /* Opening the log file */

  logstream = fopen_ctrl (logname, "w");
  fprintf (logstream, "THIS IS THE LOG OF PROGRAM \"CUBETRANSPORT\" (V%s).\n", VERSION);
  fprintf (logstream, "First .cube file is \"%s\".\n", cube0name);
  fprintf (logstream, "Second .cube file is \"%s\".\n", cube1name);
  fprintf (logstream, "The program aims at computing the %s transportation distance.\n", modename);
  fprintf (logstream, "Precision parameter has been set to %+.2f.\n", precision);


  /** READING .cube HEADERS **/

  /* Opening the .cube files */

  cube0stream = fopen_ctrl (cube0name, "r");
  cube1stream = fopen_ctrl (cube1name, "r");

  /* Looking for each file whether it was genereated by RECUBE: the result is stored in resp. "testr0" and "testr1" */

  fgets (thestring0, BUFSIZ, cube0stream);
  fgets (thestring1, BUFSIZ, cube1stream);
  sscanf (thestring0, "%s", thestring);
  sscanf (thestring1, "%s", thestring);
  testr0 = !strcmp (thestring0, "metacube");
  testr1 = !strcmp (thestring1, "metacube");

  /* Skipping the second line of each file */

  fgets (thestring0, BUFSIZ, cube0stream);
  fgets (thestring1, BUFSIZ, cube1stream);

  /* Initializing booleans for header comparison */

  testh = false; // "testh" will tell whether there is an unexcpected difference between the headers of the two .cube files
  testn = false; // "testn" will tell whether the grid sizes are different

  /* Reading the number of atoms */

  fgets (thestring0, BUFSIZ, cube0stream);
  fgets (thestring1, BUFSIZ, cube1stream);
  sscanf (thestring0, "%d", &natoms0);
  sscanf (thestring1, "%d", &natoms1);
  testh |= strcmp (thestring0, thestring1);

  /* Reading the size of the grid and the dimensions of the lattice: three lines are to be read for that */

  fgets (thestring0, BUFSIZ, cube0stream); // First line
  fgets (thestring1, BUFSIZ, cube1stream);
  testh |= strcmp (thestring0, thestring1);
  sscanf (thestring0, "%d%lf%lf%lf", &nnn.x, &lattice.x.x, &lattice.x.y, &lattice.x.z);
  sscanf (thestring1, "%d", &nnn1.x);
  testn |= (nnn1.x != nnn.x);
  fgets (thestring0, BUFSIZ, cube0stream); // Second line
  fgets (thestring1, BUFSIZ, cube1stream);
  testh |= strcmp (thestring0, thestring1);
  sscanf (thestring0, "%d%lf%lf%lf", &nnn.y, &lattice.y.x, &lattice.y.y, &lattice.y.z);
  sscanf (thestring1, "%d", &nnn1.y);
  testn |= (nnn1.y != nnn.y);
  fgets (thestring0, BUFSIZ, cube0stream); // Third line
  fgets (thestring1, BUFSIZ, cube1stream);
  testh |= strcmp (thestring0, thestring1);
  sscanf (thestring0, "%d%lf%lf%lf", &nnn.z, &lattice.z.x, &lattice.z.y, &lattice.z.z);
  sscanf (thestring1, "%d", &nnn1.z);
  testn |= (nnn1.z != nnn.z);
  cellvol = fabs (det (lattice)); // Computing the cell volume

  /* Skipping the lines relative to atomic nuclei */

  for (i = 0; true; ++i)
    {
      if (i >= natoms0 && i >= natoms1)
        {
          break;
        }
      if (i < natoms0)
        {
          fgets (thestring0, BUFSIZ, cube0stream);
        }
      if (i < natoms1)
        {
          fgets (thestring1, BUFSIZ, cube1stream);
        }
      if (i < natoms0 && i < natoms1)
        {
          testh |= strcmp (thestring0, thestring1);
        }
    }

  /* Checking whether the headers were compatible */

  if (testn)
    {
      error ("The grids of the two .cube files do not have the same number of nodes.\n");
    }
  if (testh)
    {
      warning ("The headers of the two .cube files are unexpectedly different. Most of the time, this is the sign of a deep anomaly which is going make the results incoherent; however, this might also be due to some irrelevant phenomenon like round-off approximation.\n");
    }


  /** READING .cube DATA **/

  /* Preparation steps for reading data */

  printf ("Reading data...\n");
  METAMALLOC (charge0, nnn);
  METAMALLOC (charge1, nnn);
  testd0n = false; // "testd0n" will tell whether some densities in the first .cube file are negative
  testd0z = false; // "testd0z" will tell whether some densities in the first .cube file are zero
  testd1n = false; // "testd1n" will tell whether some densities in the second .cube file are negative
  testd1z = false; // "testd1z" will tell whether some densities in the second .cube file are zero
  totcharge0 = 0;
  totcharge1 = 0;

  /* Reading the densities from .cube files */

  METAFOR (x, y, z, nnn)
  {
    fscanf (cube0stream, "%lf", &thedensity0);
    fscanf (cube1stream, "%lf", &thedensity1);
    testd0n |= thedensity0 < 0;
    testd1n |= thedensity1 < 0;
    testd0z |= thedensity0 == 0;
    testd1z |= thedensity1 == 0;

    /* Converting the densities into charges */

    thecharge0 = cellvol * thedensity0;
    thecharge1 = cellvol * thedensity1;
    charge0[x][y][z] = thecharge0;
    charge1[x][y][z] = thecharge1;
    totcharge0 += thecharge0;
    totcharge1 += thecharge1;
  }
  ENDMETAFOR

  /* Checking whether all densities were positive */

  if (testd0n || testd0z)
    {
      warning ("There were some %s densities in file \"%s\".\n", testd0n ? "negative" : "zero", cube0name);
    }
  if (testd1n || testd1z)
    {
      warning ("There were some %s densities in file \"%s\".\n", testd1n ? "negative" : "zero", cube1name);
    }
  switch (mode)
    {
    case QW:
    case LQ:
      if (testd0n || testd0z || testd1n || testd1z)
        {
          error ("The computation of the %s transportation distance can only handle positive densities.", modename);
        }
      break;
    default:
      break;
    }

  /* Balancing total charges exactly */

  totcharge = (totcharge0 + totcharge1) / 2;
  disbalanceabs = fabs (totcharge1 - totcharge0);
  disbalancerel = disbalanceabs / totcharge;
  if (disbalanceabs > DISBALANCEABS_MAX || disbalancerel > DISBALANCEREL_MAX)
    {
      warning ("The difference of total charge between the two .cube files is unexpectedly large: the charge difference is %.3f electrons, that is a relative difference of %.2f%%. I am carrying computations nonetheless, but this may lack physical relevance.\n", disbalanceabs, 100 * disbalancerel);
    }
  METAFOR (x, y, z, nnn)
  {
    charge0[x][y][z] *= totcharge / totcharge0;
    charge1[x][y][z] *= totcharge / totcharge1;
  }
  ENDMETAFOR

  /* Converting data to volumic densities, if needed */

  fprintf (logstream, "Data in file \"%s\" will be interpreted as %s densities.\n", cube0name, testr0 ? "volumic" : "punctual");
  fprintf (logstream, "Data in file \"%s\" will be interpreted as %s densities.\n", cube1name, testr1 ? "volumic" : "punctual");
  if (!testr0 || !testr1)
    {
      printf ("Converting punctual density data to volumic paradigm...\n");
      if (!testr0)
        {
          recube (&charge0, nnn);
        }
      if (!testr1)
        {
          recube (&charge1, nnn);
        }
    }


  /** COMPUTING THE TRANSPORTATION DISTANCE **/

  printf ("Computing the transportation distance...\n");
  switch (mode)
    {
    case TV:
      distance = tvdistance (charge0, charge1);
      break;
    case KR:
      distance = krdistance (charge0, charge1);
      break;
    case QW:
      distance = qwdistance (charge0, charge1);
      break;
    case LQ:
      distance = lqdistance (charge0, charge1);
      break;
    }


  /** EXITING **/

  printf ("Computation over.\n");
  ffprintf (2, stdout, logstream, "The %s transportation distance between the densities worths %f a.u..\n", modename, distance);
  METAFREE (nnn, charge0);
  METAFREE (nnn, charge1);
  return EXIT_SUCCESS;
}
