#include "cubetransport.h"
#include "quadratic.h"

double qwdistance (double ***charge0, double ***charge1)
{


  /** VARIABLE DECLARATIONS **/

  /* Data meta-arrays, also indexed by time */

  double ****charget; // Charges along time and space
  double (****flowt)[3]; // Transportation plan along time, expressed as a sequence of flow structures

  /* Important variables */

  double *p_subchargea[T1_MAX + 1], *p_subchargeb[T1_MAX + 1]; // Charge along time at the start point A of the zone interesting us, resp. at the finish point B
  double *p_subflow [T1_MAX]; // Charge transported from A to B along time
  struct double3 subdipolea[T1_MAX], subdipoleb[T1_MAX]; // Flow current (expressed as a dipole) around point A, resp. around point B, along time
  double cost; // Transportation cost
  double distance, lastvalues[3]; // Transportation distance (square root of cost), and last distances computed
  double zero_descent; // Threshold to halt the gradient descent
  int iteration; // Iteration number

  /* Indices and auxiliary variables */

  int t; // Index for time
  int x, y, z, xb, yb, zb; // (x,y,z) are indices for locating a point in the grid, in particular for point A; (xb,yb,zb) are for locating point B
  int v; // Index for a flow direction
  double fx, fy, fz; // Transported charge along the x direction, resp. along y, resp. along z


  /** PREPARING STUFF FOR THE GRADIENT DESCENT **/

  /* Setting "zero_descent" and "t1" */

  t1 = (int) fmin (ceil (T1_DEFAULT * pow (2, precision / 2)), (double) T1_MAX);
  zero_descent = ZERO_DESCENT_DEFAULT / pow (2, precision);
  fprintf (logstream, "The number of time discretization steps for the time-continuous transportation plan will be %d.\n", t1);
  fprintf (logstream, "The threshold to consider stationarity as attained has been set at %.2E a.u..\n", zero_descent);

  /* Initializing "charget" */

  printf ("Initializing the transportation plan to optimize...\n");
  charget = malloc ((t1 + 1) * sizeof(double ***));
  for (t = 0; t <= t1; ++t)
    {
      METAMALLOC (charget[t], nnn);
    }
  for (t = 0; t <= t1; ++t)
    {
      METAFOR (x, y, z, nnn)
      {
        charget[t][x][y][z] = charge0[x][y][z] + (double)t / t1 * (charge1[x][y][z] - charge0[x][y][z]);
      }
      ENDMETAFOR
    }

  /* Initializing "flowt" */

  flowt = malloc (t1 * sizeof(double (***)[3]));
  for (t = 0; t < t1; ++t)
    {
      METAMALLOC (flowt[t], nnn);
      initflow (charget[t], charget[t + 1], flowt[t]);
    }


  /** GENERAL GRADIENT DESCENT STRUCTURE **/

  ffprintf (2, stdout, logstream, "Iterating for the minimal transportation cost...\n");
  for (iteration = 0; true; ++iteration)
    {

      /* We compute the transportation distance at every "round" iteration */

      if (isround (iteration))
        {
          cost = 0;
          for (t = 0; t < t1; ++t)
            {
              cost += qcost (flowt[t], charget[t], charget[t + 1]);
            }
          cost *= t1;
          distance = sqrt (cost);
          ffprintf (2, stdout, logstream, "Iteration #%d: %f a.u..\n", iteration, distance);
          fflush (logstream);

          /* Checking if stationarity has been attained */

          lastvalues[2] = lastvalues[1];
          lastvalues[1] = lastvalues[0];
          lastvalues[0] = distance;
          if (lastvalues[2] - lastvalues[0] < zero_descent && iteration >= ITERATION_MIN)
            {
              ffprintf (2, stdout, logstream, "Stationarity has been attained!\n");
              break;
            }
        }


      /** DESCENT ALONG SPACE **/

      for (t = 0; t < t1; ++t)
        {
          descent (flowt[t], charget[t], charget[t + 1]);
        }


      /** DESCENT ALONG TIME **/

      METAFOR (x, y, z, nnn)
      {
        for (v = 0; v < 3; ++v)
          {

            /* Setting "xb", "yb" and "zb" */

            xb = x;
            yb = y;
            zb = z;
            switch (v)
              {
              case X:
                xb += 1;
                break;
              case Y:
                yb += 1;
                break;
              case Z:
                zb += 1;
                break;
              }
            if (xb < nnn.x && yb < nnn.y && zb < nnn.z)
              {

                /* Storing concerned data in dedicated arrays */

                for (t = 0; t <= t1; ++t)
                  {
                    p_subchargea[t] = &charget[t][x][y][z];
                    p_subchargeb[t] = &charget[t][xb][yb][zb];
                  }
                for (t = 0; t < t1; ++t)
                  {
                    p_subflow[t] = &flowt[t][x][y][z][v];
                    fx = (flowat(flowt[t], x - 1, y, z, X) + flowat(flowt[t], x, y, z, X)) / 2;
                    fy = (flowat(flowt[t], x, y - 1, z, Y) + flowat(flowt[t], x, y, z, Y)) / 2;
                    fz = (flowat(flowt[t], x, y, z - 1, Z) + flowat(flowt[t], x, y, z, Z)) / 2;
                    subdipolea[t] = lincomb (fx, fy, fz);
                    fx = (flowat(flowt[t], xb - 1, yb, zb, X) + flowat(flowt[t], xb, yb, zb, X)) / 2;
                    fy = (flowat(flowt[t], xb, yb - 1, zb, Y) + flowat(flowt[t], xb, yb, zb, Y)) / 2;
                    fz = (flowat(flowt[t], xb, yb, zb - 1, Z) + flowat(flowt[t], xb, yb, zb, Z)) / 2;
                    subdipoleb[t] = lincomb (fx, fy, fz);
                  }

                /* Performing time descent step */

                codescent (p_subflow, p_subchargea, p_subchargeb, subdipolea, subdipoleb, v);
              }
          }
      }
      ENDMETAFOR
    }

  /** EXITING THE FUNCTION **/


  for (t = 0; t <= t1; ++t)
    {
      METAFREE (nnn, charget[t]);
    }
  free (charget);
  for (t = 0; t < t1; ++t)
    {
      METAFREE (nnn, flowt[t]);
    }
  free (flowt);
  return distance;
}
